#include <iostream>
#include <fstream>
#include "inc/Geodesic.h"
#include "smreadopener.h"
#include <hash_map>


class vec{
public:
	float p[3];

	vec(){};

	vec(float* v){
		for(int i=0;i<3;i++)
			p[i]=v[i];
	}

	vec operator-(vec v){
		vec ret;
		for(int i=0;i<3;i++)
			ret.p[i]=p[i]-v.p[i];
		return ret;
	}

	float len2(){
		return p[0]*p[0]+p[1]*p[1]+p[2]*p[2];
	}
};

stdext::hash_map<int,vec> vertex_hash;

bool goodTriangle(int *t_idx, float len){
	if (len == 0)
		return true;
	vec p[3];
	for(int i=0;i<3;i++)
		p[i]=vertex_hash.find(t_idx[i])->second;

	if (((p[0]-p[1]).len2() > len) || (((p[0]-p[2]).len2() > len) || ((p[1]-p[2]).len2() > len)))
		return false;
	else
		return true;
}

void loadOBJ(const char* file, Geodesic& g)
{
    g.meshClear();

    using namespace std;

    ifstream is(file);
    printf("Reading \'%s\' ... ", file);
    fflush(stdout);

    while (is) {
        char str[1024];
        is.getline(str, 1024);
        bool verticesDone = false;

        switch (str[0]) {
            case 'v':
                if (isspace(str[1]) && !verticesDone) { // 'vt', 'vp', 'vn' are ignored
                    double x, y, z;
                    int res = sscanf(&str[2], "%lf %lf %lf", &x, &y, &z);
                    if (res == 3) // for inputs with 'v x y z w', w is ignored
                        g.meshAddVertex(x, y, z);
                }
                break;
            case 'f':
                const int n = g.meshNVertices();
                int i, j, k;
                // for inputs with n-gons only a triangle for every n-gon is built
                int res = sscanf(str, "f %i %i %i", &i, &j, &k);
                if (res == 3 && 1 <= i && i <= n && 1 <= j && j <= n && 1 <= k && k <= n)
                {
                    g.meshAddFace(i-1, j-1, k-1);
                }
                break;
        }
    }

    printf("Done\n");
    fflush(stdout);
}

void loadSM(const char* file, Geodesic& g, float zscale, float maxlen)
{
    g.meshClear();

    using namespace std;

	SMreadOpener* smreadopener = new SMreadOpener();
	smreadopener->set_file_name(file);

	SMreader* smreader=0;
	if (!(smreader=smreadopener->open())){
		cout<<"Could not open "<<file<<endl;
		exit(1);
	}



	SMevent sm_event;
    printf("Reading \'%s\' ... ", file);
    fflush(stdout);

    while (sm_event = smreader->read_element()) {

        switch (sm_event) {
            case SM_VERTEX:
                g.meshAddVertex(smreader->v_pos_f[0],smreader->v_pos_f[1],zscale*smreader->v_pos_f[2]);
				vertex_hash.insert(make_pair(smreader->v_count-1, vec(smreader->v_pos_f)));
                break;
            case SM_TRIANGLE:
				if (goodTriangle(smreader->t_idx,maxlen))
					g.meshAddFace(smreader->t_idx[0], smreader->t_idx[1], smreader->t_idx[2]);
				else
					cout<<"discarded triangle "<<smreader->t_idx[0]<<" "<<smreader->t_idx[1]<<" "<<smreader->t_idx[2]<<endl;
				for(int i=0;i<3;i++)
					if (smreader->t_final[i])
						vertex_hash.erase(smreader->t_idx[i]);
				break;
			case SM_FINALIZED:
				vertex_hash.erase(smreader->final_idx);
			default:
				break;
        }
    }

	if (!(vertex_hash.empty())){
		cout<<"Some vertices have not been finalized"<<endl;
		vertex_hash.clear();
	}
	smreadopener->close();
	delete smreader;

    printf("Done\n");
    fflush(stdout);
}


void savePath(Geodesic& g, int s, int t, const char* meshFile, int num_vert)
{
	std::cout<<"Computation finished"<<std::endl;
    char str[1024];
    sprintf(str, "%s.%d.%d.path", meshFile, s+1, t+1);
    std::ofstream os(str);
	float *points=(float *)malloc(num_vert*3*sizeof(float));

	SMreadOpener* smreadopener = new SMreadOpener();
	smreadopener->set_file_name(meshFile);

	SMreader* smreader=smreadopener->open();

	SMevent sm_event;

	int i;

	while (sm_event = smreader->read_element()) {
        switch (sm_event) {
            case SM_VERTEX:
				for(i=0;i<3;i++)
                    points[3*smreader->v_count-3+i]=smreader->v_pos_f[i];
                break;
 			default:
				break;
        }
    }
	smreadopener->close();
	delete smreader;

	float point[3];

	for(i=0;i<3;i++)
		point[i]=points[3*s+i];

	bool start=false;
	int num_pts=0;
    while (!g.pathEmpty()) {
		int vs, vt;
		double b;
		g.pathPopNextStep(vs, vt, b);
		num_pts++;
	//	std::cout<<vs<<" "<<vt<<" "<<b<<std::endl;
		if (b==1)
			b=0;
		else if (b==0)
			b=1;
		for(i=0;i<3;i++)
			point[i]=(1-b)*points[3*vs+i]+b*points[3*vt+i];

		sprintf(str, "%f %f %f\n", point[0], point[1], point[2]);
	    os << str;
	}
		std::cout<<"Number of points in path: "<<num_pts<<std::endl;

	//	if(start){
	//        sprintf(str, "%f %f %f\n", point[0], point[1], point[2]);
	//	    os << str;

	//		for(i=0;i<3;i++)
	//			point[i]=(1-b)*points[3*vs+i]+b*points[3*vt+i];
	//	}
	//	else
	//		start=true;
 //   }
	//sprintf(str, "%f %f %f\n", points[3*t+0], points[3*t+1], points[3*t+2]);
	//os << str;
	free(points);
}

void saveDistances(Geodesic& g, int s, const char* meshFile)
{
	std::cout<<"Computation finished"<<std::endl;
    char str[1024];
    sprintf(str, "%s.%d.dist", meshFile, s);
    std::ofstream os(str);

    const int n = g.meshNVertices();
    for (int i = 0; i < n; ++i) {
        const double d = g.distanceAtVertex(i);
        sprintf(str, "%.14g\n", d);
        os << str;
    }
}

void printUsage()
{
    printf("Usage:\n\n");
    printf("GeoTest.exe mesh.obj -s VID [-d] [-a Tol] [-t VID] [-p0 VID] [-p1 VID]\n\n");
    printf("-s VID: specify the source vertex VID;\n");
    printf("  Note: if -t is not specifed the geodesics are computed for the entire mesh;\n");
    printf("-t VID: find the shortest path between the source and target (VID) vertices only;\n");
    printf("-d: output the distance into a file;\n");
    printf("-a Tol: run approximate algorithm given tolerance, 0 < Tol < 1;\n");
    printf("-p0 VID: find the shortest path between the source vertex and VID;\n");
    printf("-p1 VID: find the shortest path between the source vertex and VID;\n");
}

int main(int argc, char *argv[])
{
    printf("Version: 1.0\n");
    fflush(stdout);

    if (argc < 2) {
        printUsage();
        return 0;
    }

    const char* meshFile = 0;
    int sourceVertexID = -1;
    int targetVertexID = -1;
    int pathToVertexIDs[2] = { -1, -1 };
    bool printDistances = false;
    double tolerance = -1.0;
	float zscale=1.0f;
	float maxlen=0.0f;

    for (int i = 1; i < argc; ++i) {
        if (argv[i][0] == '-') {
            switch (argv[i][1]) {
            case 's':
                {
                    int id;
                    int res = sscanf(argv[i+1], "%d", &id);
                    if (res == 1 && i >= 0) {
                        sourceVertexID = id;
                        ++i;
                    }
                    else {
                        printf("Error: -s option requires numeric parameter. \n");
                        printUsage();
                        return 0;
                    }
                }
                break;
            case 't':
                {
                    int id;
                    int res = sscanf(argv[i+1], "%d", &id);
                    if (res == 1 && i >= 0) {
                        targetVertexID = id;
                        ++i;
                    }
                    else {
                        printf("Error: -t option requires numeric parameter. \n");
                        printUsage();
                        return 0;
                    }
                }
                break;
            case 'd':
                {
                    printDistances = true;
                }
                break;
			case 'l':
                {
                    maxlen=atof(argv[++i]);
                }
                break;
            case 'a':
                {
                    if (i + 1 == argc) {
                        printf("Error: -a option requires a valid tolerance parameter. \n");
                        printUsage();
                        return 0;
                    }
                    double tol;
                    int res = sscanf(argv[i+1], "%lf", &tol);
                    if (res == 1 && 0.0 < tol && tol < 1.0) {
                        tolerance = tol;
                        ++i;
                    }
                    else {
                        printf("Error: -a option requires a valid tolerance parameter. \n");
                        printUsage();
                        return 0;
                    }
                }
                break;
            case 'p':
                {
                    int id;
                    int res = sscanf(argv[i+1], "%d", &id);
                    if ((argv[i][2] == '0' || argv[i][2] == '1') && res == 1 && i >= 0) {
                        if (argv[i][2] == '0')
                            pathToVertexIDs[0] = id;
                        else if (argv[i][2] == '1')
                            pathToVertexIDs[1] = id;
                        ++i;
                    }
                    else {
                        printf("Error: -pi option requires numeric parameter. \n");
                        printUsage();
                        return 0;
                    }
                }
                break;
			case 'z':
                {
                   zscale=atof(argv[++i]);
                }
                break;
            default:
                {
                    printf("Warning: Option %s has been ignored.\n", argv[i]);
                }
            }
        }
        else
            meshFile = argv[i];
    }

    if (meshFile == 0) {
        printf("Error: No mesh file has been specified. \n");
        printUsage();
        return 0;
    }

    Geodesic geodesic;

    loadSM(meshFile, geodesic, zscale, maxlen*maxlen);

    if (geodesic.meshNFaces() == 0) {
        printf("Error reading the mesh file '%s' or the input mesh does not have any faces.\n", meshFile);
        return 0;
    }
    else {
        printf("The mesh in '%s' has %d vertices and %d faces.\n",
            meshFile, geodesic.meshNVertices(), geodesic.meshNFaces());
    }

	int num_vert=geodesic.meshNVertices();
    if (sourceVertexID == -1) {
        printf("Error: The source vertex has not been specified.\n");
        printUsage();
        return 0;
   }

    if (targetVertexID != -1) {
        geodesic.findPathBetweenTwoVertices(sourceVertexID-1, targetVertexID-1);
		std::cout<<"Path computed"<<std::endl;
        savePath(geodesic, sourceVertexID-1, targetVertexID-1, meshFile, num_vert);
		char tchar;
		std::cout<<"Press any key and return to exit:"<<std::endl;
		std::cin>>tchar;
        return 1;
    }

    if (tolerance == -1.0)
        geodesic.computeExactGeodesics(sourceVertexID-1);
    else
        geodesic.computeApproxGeodesics(sourceVertexID-1, tolerance);

    if (pathToVertexIDs[0] != -1) {
        geodesic.findPathToVertex(pathToVertexIDs[0]-1);
        savePath(geodesic, sourceVertexID-1, pathToVertexIDs[0]-1, meshFile, num_vert);
    }

    if (pathToVertexIDs[1] != -1) {
        geodesic.findPathToVertex(pathToVertexIDs[1]-1);
        savePath(geodesic, sourceVertexID-1, pathToVertexIDs[1]-1, meshFile, num_vert);
    }

    if (printDistances)
        saveDistances(geodesic, sourceVertexID-1, meshFile);

	char tchar;
	std::cout<<"Press any key and return to exit:"<<std::endl;
	std::cin>>tchar;

    return 1;
}